/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2004 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "ButtonIcon.h"

ButtonIcon::ButtonIcon(const string &name)
{
	defaulticon.iconSize = GetxRCInt(name, "IconSize", 16, 8, 64);
	defaulticon.iconX = GetxRCString(name, "IconX", "0");
	defaulticon.iconY = GetxRCString(name, "IconY", "" );
	
	defaulticon.icon_saturation = GetxRCInt(name, "IconSaturationIntensity", 255, 0, 255);
	defaulticon.icon_hueintensity = GetxRCInt(name, "IconHueIntensity", 0, 0, 255);
	defaulticon.icon_huecolor = GetxRCColor(name, "IconHueColor", RGB(255, 255, 255));
}

ButtonIcon::~ButtonIcon()
{
}

void ButtonIcon::subconfigure(string type, ICONSETTINGS &workSettings, ICONSETTINGS &defaultSettings)
{
	workSettings.iconSize = GetxRCInt(type, "IconSize", defaultSettings.iconSize, 8, 64);
	workSettings.iconX = GetxRCString(type, "IconX", defaultSettings.iconX);
	if ( defaulticon.iconY.empty() )
		workSettings.iconY = GetxRCString(type, "IconY", ("-" + IntToString(workSettings.iconSize/2) + "c"));
	else
		workSettings.iconY = GetxRCString(type, "IconY", defaulticon.iconY);

	workSettings.icon_saturation = GetxRCInt(type, "IconSaturationIntensity", defaultSettings.icon_saturation, 0, 255);
	workSettings.icon_hueintensity = GetxRCInt(type, "IconHueIntensity", defaultSettings.icon_hueintensity, 0, 255);
	workSettings.icon_huecolor = GetxRCColor(type, "IconHueColor", defaultSettings.icon_huecolor);
}

void ButtonIcon::configure(const string &name)
{
	subconfigure((name + "Normal"), normal, defaulticon);

	subconfigure((name + "Active"), active, normal);

	subconfigure((name + "Minimized"), minimized, normal);

	subconfigure((name + "Flashing"), flashing, normal);

	subconfigure((name + "Hover"), hover, defaulticon);

	subconfigure((name + "Grouped"), group, defaulticon);
}

void ButtonIcon::apply(HDC hDC, int x, int y, int width, int height, int buttonType, HICON hIcon, bool useAlphaMap)
{
	if(hIcon == NULL)
		return;

	int iconX = 0;
	int iconY = 0;
	int iconSize = 16;

	int _saturation = 255;
	int _hueIntensity = 0;
	COLORREF _huecolor = RGB(255, 255, 255);

	if (buttonType == 0)
	{
		iconX = xParseCoord(normal.iconX, 0, width);
		iconY = xParseCoord(normal.iconY, 0, height);

		iconSize = normal.iconSize;
		_saturation = normal.icon_saturation;
		_hueIntensity = normal.icon_hueintensity;
		_huecolor = normal.icon_huecolor;
	}
	else if (buttonType == 1)
	{
		iconX = xParseCoord(active.iconX, 0, width);
		iconY = xParseCoord(active.iconY, 0, height);

		iconSize = active.iconSize;
		_saturation = active.icon_saturation;
		_hueIntensity = active.icon_hueintensity;
		_huecolor = active.icon_huecolor;
	}
	else if (buttonType == 2)
	{
		iconX = xParseCoord(minimized.iconX, 0, width);
		iconY = xParseCoord(minimized.iconY, 0, height);

		iconSize = minimized.iconSize;
		_saturation = minimized.icon_saturation;
		_hueIntensity = minimized.icon_hueintensity;
		_huecolor = minimized.icon_huecolor;
	}
	else if (buttonType == 3)
	{
		iconX = xParseCoord(flashing.iconX, 0, width);
		iconY = xParseCoord(flashing.iconY, 0, height);

		iconSize = flashing.iconSize;
		_saturation = flashing.icon_saturation;
		_hueIntensity = flashing.icon_hueintensity;
		_huecolor = flashing.icon_huecolor;
	}
	else if (buttonType == 4)
	{
		iconX = xParseCoord(hover.iconX, 0, width);
		iconY = xParseCoord(hover.iconY, 0, height);

		iconSize = hover.iconSize;
		_saturation = hover.icon_saturation;
		_hueIntensity = hover.icon_hueintensity;
		_huecolor = hover.icon_huecolor;
	}
	else if (buttonType == 5)
	{
		iconX = xParseCoord(group.iconX, 0, width);
		iconY = xParseCoord(group.iconY, 0, height);

		iconSize = group.iconSize;
		_saturation = group.icon_saturation;
		_hueIntensity = group.icon_hueintensity;
		_huecolor = group.icon_huecolor;
	}

	if ( useAlphaMap || _saturation != 255 || _hueIntensity != 0 )
	{
		HDC hdcBGCOPY = CreateCompatibleDC(hDC);
		HBITMAP hbmBGCOPY = CreateCompatibleBitmap(hDC, iconSize, iconSize);
		SelectObject(hdcBGCOPY, hbmBGCOPY);

		//Copy RealBG in Result Background
		BitBlt( hdcBGCOPY, 0, 0, iconSize, iconSize, hDC, x+iconX, y+iconY, SRCCOPY );

		HDC hdcBuffer = CreateCompatibleDC(hDC);
		HBITMAP hbmBuffer = CreateCompatibleBitmap(hDC, iconSize, iconSize);
		SelectObject(hdcBuffer, hbmBuffer);

		//Copy RealBG in Result Background
		BitBlt( hdcBuffer, 0, 0, iconSize, iconSize, hdcBGCOPY, 0, 0, SRCCOPY );

		//Draw Icon
		DrawIconEx(hdcBuffer, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_IMAGE);

		HDC hdcMask = CreateCompatibleDC(hDC);
		HBITMAP hbmMask = CreateCompatibleBitmap(hDC, iconSize, iconSize);
		SelectObject(hdcMask, hbmMask);

		//Draw Mask of the Icon
		DrawIconEx(hdcMask, 0, 0, hIcon, iconSize, iconSize, 0, NULL, DI_MASK);

		int iy = 0;
		int ix = 0;
		COLORREF cl = 0;
		BYTE r = 0;
		BYTE g = 0;
		BYTE b = 0;
		for(iy = 0; iy < iconSize; iy++)
		{
			for(ix = 0; ix < iconSize; ix++)
			{
				// Loop through all transparent pixels...
				// If the mask-pixel is white, then break
				if(!GetPixel(hdcMask, ix, iy))
				{
					cl = GetPixel(hdcBuffer, ix, iy);
					r = GetRValue(cl);
					g = GetGValue(cl);
					b = GetBValue(cl);

					// Saturation effect
					if(_saturation != 255)
					{
						BYTE gray = (BYTE)(r * 0.3086 + g * 0.6094 + b * 0.0820);

						r = (BYTE)((r * _saturation + gray * (255 - _saturation) + 255) >> 8);
						g = (BYTE)((g * _saturation + gray * (255 - _saturation) + 255) >> 8);
						b = (BYTE)((b * _saturation + gray * (255 - _saturation) + 255) >> 8);
					}

					// Hue effect
					if(_hueIntensity != 0 )
					{
						// The B & R values of the hue is swapped here, can somebody tell me why it only works that way?
						// test normal
						r = (BYTE)((r * (255 - _hueIntensity) + GetRValue(_huecolor) * _hueIntensity + 255) >> 8);
						g = (BYTE)((g * (255 - _hueIntensity) + GetGValue(_huecolor) * _hueIntensity + 255) >> 8);
						b = (BYTE)((b * (255 - _hueIntensity) + GetBValue(_huecolor) * _hueIntensity + 255) >> 8);
					}

					SetPixelV(hdcBGCOPY, ix, iy, RGB(r, g, b));
				}
			}
		}

		BitBlt( hDC, x+iconX, y+iconY, iconSize, iconSize, hdcBGCOPY, 0, 0, SRCCOPY);
				
		hbmMask = (HBITMAP) SelectObject(hdcMask, hbmMask);
		DeleteObject(hbmMask);
		DeleteDC(hdcMask);

		hbmBuffer = (HBITMAP) SelectObject(hdcBuffer, hbmBuffer);
		DeleteObject(hbmBuffer);
		DeleteDC(hdcBuffer);

		hbmBGCOPY = (HBITMAP) SelectObject(hdcBGCOPY, hbmBGCOPY);
		DeleteObject(hbmBGCOPY);
		DeleteDC(hdcBGCOPY);
	}
	else
		DrawIconEx(hDC, x+iconX, y+iconY, hIcon, iconSize, iconSize, 0, NULL, DI_NORMAL);
}

