/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2004 The LS-Universe Team

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#include "stdafx.h"
#include "Taskbar.h"
#include "TaskbarButton.h"
#include "TaskbarTexture.h"
#include "ButtonTexture.h"
#include "ButtonIcon.h"
#include "ButtonFont.h"

#include "DropTarget.h"
#include "WindowUtilities.h"

#include "shlobj.h"
#include <algorithm>

#define LM_UPDATEBG (WM_USER + 1)

#define GET_X_LPARAM(lp) ((int)(short)LOWORD(lp))
#define GET_Y_LPARAM(lp) ((int)(short)HIWORD(lp))

//hack!!
int nButtonsAddedAtStartup;

extern int screenX;
extern int screenY;

extern int bitsperpixel;

// undocumented SwitchToThisWindow API
void (WINAPI *SwitchToThisWindow)(HWND, int);

//Alpha Transparency
typedef DWORD (WINAPI *PSLWA)(HWND, DWORD, BYTE, DWORD);
PSLWA pSetLayeredWindowAttributes = NULL;

typedef BOOL (WINAPI *PULW)(HWND hwnd, HDC hdcDst, POINT *pptDst, SIZE *psize, HDC hdcSrc, POINT *pptSrc, COLORREF crKey, BLENDFUNCTION *pblend, DWORD dwFlags);
PULW pUpdateLayeredWindow = NULL;

#define TIMER_UPDATE		1
#define TIMER_MOUSETRACK	2
#define TIMER_ALPHAFADEIN	3
#define TIMER_ALPHAFADEOUT	4
#define TIMER_DROPTIMEOUT	5

UINT lsMessages[] = {
                      LM_LSSELECT,
                      LM_REDRAW,
                      LM_WINDOWACTIVATED,
                      LM_WINDOWCREATED,
                      LM_WINDOWDESTROYED,
					  LM_REFRESH,
                      0
					 };


//////////////////////////////////////////////////////////////////////
// Konstruktion/Destruktion
//////////////////////////////////////////////////////////////////////

Taskbar::Taskbar(const string &name):actualx(0),actualy(0),actualwidth(0),actualheight(0)
{
	this->name = name;

	hWnd = 0;
	hInstance = 0;
	box = 0;

	backupExportedButtonCount = 0;

	createLayoutLocked = true;
	addLock = -1;
	removeLock = -1;

	afterClass = "";
	afterTitle = "";

	lines_rows = 1;
	lines_rowsAdded = 0;

	wrapCount = -1;

	backgroundDC = 0;
	backgroundBitmap = 0;

	alphaInitialized = false;
	layeredFlag = false;

	alphaFade = false;
	UpdateAlpha = 0;
	alphaFadeStep = 2;
	alphaFadeDelay = 10;

	autoSort = 0;

	mouseInside = false;
	focusOnEnter = 0;
	actionsOnDownButton = false;

	lastActive = NULL;

	visible = false;
	autoHidden = false;

	inCriticalProcess = false;

	minimizeAtAllCosts = false;

	taskbartexture = new TaskbarTexture();
	buttontexture = new ButtonTexture(name);
	buttonicon = new ButtonIcon(name);
	buttonfont = new ButtonFont(name);

	movecapture = NULL;
	capture = NULL;
	clickcapture = NULL;

	SwitchToThisWindow = (void (WINAPI *)(HWND, int)) 
							GetProcAddress(GetModuleHandle("USER32.DLL"), "SwitchToThisWindow");

	pDropTarget = NULL;
	dropcapture = NULL;

	displayedtaskbuttons.clear();
	taskbuttons.clear();
}

Taskbar::~Taskbar()
{
	RemoveList.clear();
	CustomIconList.clear();

	AddMatchingList.clear();
	RemoveMatchingList.clear();

	if (pDropTarget)
	{
		pDropTarget->Revoke( hWnd );
		delete pDropTarget;
	}

	if (hWnd && ToolTips)
	{
		updateHint(hWnd,"", false, true);
		DestroyWindow(ToolTips);
		ToolTips = 0;
	}

	// shutdown timer
	if(IsWindow(hWnd))
	{
		KillTimer(hWnd, TIMER_UPDATE);
		KillTimer(hWnd, TIMER_MOUSETRACK);
		KillTimer(hWnd, TIMER_ALPHAFADEIN);
		KillTimer(hWnd, TIMER_ALPHAFADEOUT);
		KillTimer(hWnd, TIMER_DROPTIMEOUT);
	}

	displayedtaskbuttons.clear();
	// release buttons
	for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
	{	
		delete (*tbit);
		*tbit = NULL;
	}
	taskbuttons.clear();

	if(IsWindow(hWnd))
	{
		DestroyWindow(hWnd);
	}

	if(backgroundBitmap)
	{
		backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
		DeleteObject(backgroundBitmap);
	}
	if (backgroundDC)
		DeleteDC(backgroundDC);

	if (buttontexture)
		delete buttontexture;
	if (buttonicon)
		delete buttonicon;
	if (buttonfont)
		delete buttonfont;
	if (taskbartexture)
		delete taskbartexture;

	// unregister messages
	SendMessage( GetLitestepWnd(), LM_UNREGISTERMESSAGE, (WPARAM) hWnd, (LPARAM) ::lsMessages );

	if ( VWMCurrentDesktopOnly )
	{
		UINT uMsgs[] = { 9370, 0 };

		SendMessage( GetLitestepWnd(), LM_UNREGISTERMESSAGE ,(WPARAM) hWnd, (LPARAM) uMsgs );
	}
}

void Taskbar::load(HINSTANCE hInstance, HWND box)
{
	this->hInstance = hInstance;
	
	if (IsWindow(box))
	{
		this->box = box;
	}
	else
	{
		this->box = NULL;
	}
	
	if (IsWindow(hWnd))	// window already exists...
		return;

	hWnd = CreateWindowEx(box ? 0 : WS_EX_TOOLWINDOW,
		"xTaskbar",
		name.c_str(),
		(box ? WS_CHILD : WS_POPUP),
		actualx, actualy,		
		actualwidth, actualheight,
		box ? box : GetLitestepWnd(), 
		0,
		hInstance,
		this);
		
	if (hWnd)
	{
		SendMessage( GetLitestepWnd(), LM_REGISTERMESSAGE, (WPARAM) hWnd, (LPARAM) ::lsMessages );
		
		SetWindowLong(hWnd, GWL_USERDATA, magicDWord);

		reconfigure();

		if ( VWMCurrentDesktopOnly )
		{
			UINT uMsgs[] = { 9370, 0 };

			SendMessage( GetLitestepWnd(), LM_REGISTERMESSAGE ,(WPARAM) hWnd, (LPARAM) uMsgs );
		}

		if( nButtonsAddedAtStartup > 0 )
		{
			for(int nI = 0; nI < nButtonsAddedAtStartup; ++nI)
			{
				if( !taskbarAdd.empty() && displayButtonType < 0 )
					LSExecute(hWnd, taskbarAdd.c_str(), SW_SHOWNORMAL);
			}
			if ( wrapCount != -1 )
			{
				int addLines = 0;
				int tmpcounter = 1;
				while ( tmpcounter <= nButtonsAddedAtStartup )
				{
					if ( tmpcounter/wrapCount >= lines_rows && tmpcounter%wrapCount == 1 )
						addLines++;
					tmpcounter++;
				}

				while ( addLines > 0 )
				{
					lines_rowsAdded++;
					if ( !taskbarLineAdd.empty() )
						LSExecute(hWnd, taskbarLineAdd.c_str(), SW_SHOWNORMAL);
					addLines--;
				}
			}
			addLock = nButtonsAddedAtStartup;

			backupExportedButtonCount = nButtonsAddedAtStartup;

			nButtonsAddedAtStartup = 0;
		}

		createLayoutLocked = false;
		
		CreateLayout();

		SetTimer(hWnd, TIMER_UPDATE, timerRefresh, NULL);

		if ( !hideIfEmpty || !autoHidden )
		{
			if (startHidden || loadInBox)
				hide();
			else
				show();
		}

		//we become a drop target
		pDropTarget = new DropTarget;
		pDropTarget->Register( hWnd );
		pDropTarget->RegisterDrop( hWnd, MOUSEISDRAGGING );
	}
	else
		delete this;
}

void Taskbar::reconfigure(bool refresh)
{
	nButtonsAddedAtStartup = 0;

	NameValuePair buttonActionValues[] = {
			{ ".none", 0 },
			{ ".select", 1 },
			{ ".menu", 2 },
			{ ".close", 3 },
			{ ".ontop", 4 },
			{ 0, 0 }
			};

	NameValuePair directionValues[] = {
			{ ".right", 1 },
			{ ".left", 2 },
			{ ".bottom", 3 },
			{ ".top", 4 },
			{ 0, 0 }
			};

	NameValuePair wrapdirectionValues[] = {
			{ ".bottom", 1 },
			{ ".top", 2 },
			{ ".right", 3 },
			{ ".left", 4 },
			{ 0, 0 }
			};

	NameValuePair displayValues[] = {
			{ "flashing", -2},
			{ "all", -1 },
			{ "normal+active+minimized", -1 },
			{ "normal", 0 },
			{ "active", 1 },
			{ "minimized", 2 },
			{ "active+normal", 3 },
			{ "normal+active", 3 },
			{ "minimized+normal", 4 },
			{ "normal+minimized", 4 },
			{ "active+minimized", 5 },
			{ "minimized+active", 5 },
			{ 0, 0 }
			};

	NameValuePair layoutValues[] = {
			{ ".horizontal", 0 },
			{ ".vertical", 1 },
			{ 0, 0 }
			};


	NameValuePair autoSortValues[] = {
			{ ".none", 0 },
			{ ".filename", 1 },
			{ ".classname", 2 },
			{ ".windowname", 3 },
			{ 0, 0 }
			};

	RemoveList.clear();
	CustomIconList.clear();

	AddMatchingList.clear();
	RemoveMatchingList.clear();

	GetxRCList("*xTaskbarRemove", RemoveList, true);
	GetxRCList("*xTaskbarIcon", CustomIconList, false);

	AddMatchingList = split( GetxRCString(name, "AddOnlyMatching", ""), ";");
	RemoveMatchingList = split( GetxRCString(name, "RemoveOnlyMatching", ""), ";");

	char classname[1024], windowname[1024];
	char *zorderbuffers[] = {classname, windowname};

	if ( LCTokenize( GetxRCLine(name, "BehindWindow", "").c_str(), zorderbuffers, 2, NULL) >= 1 )
	{
		afterClass = classname;
		afterTitle = windowname;
	}

	hideIfEmpty = GetxRCBool(name, "HideIfEmpty");

	autoSort = GetxRCNamedValue(name, "Sort", autoSortValues, 0);

	displayButtonType = GetxRCNamedValue(name, "Display", displayValues, -1);

	actionsOnDownButton = GetxRCBool(name, "ActionsOnButtonDown");

	lButtonAction = GetxRCNamedValue(name, "OnLeftClickButton", buttonActionValues, 1);
	rButtonAction = GetxRCNamedValue(name, "OnRightClickButton", buttonActionValues, 2);
	mButtonAction = GetxRCNamedValue(name, "OnMiddleClickButton", buttonActionValues, 0);

	useGrouping = GetxRCBool(name, "UseGrouping");

	useBigIcon = GetxRCBool(name, "UseBigIcon");

	noMinimizeOnClick = GetxRCBool(name, "NoMinimizeOnClick");

	VWMCurrentDesktopOnly = GetxRCBool(name, "VWMCurrentDesktopOnly");

	//Events/Actions
	taskbarLButtonUp = GetxRCLine(name, "OnLeftClickUp", "");
	taskbarLButtonDown =GetxRCLine(name, "OnLeftClickDown", "");
	taskbarMButtonUp = GetxRCLine(name, "OnMiddleClickUp", "");
	taskbarMButtonDown = GetxRCLine(name, "OnMiddleClickDown", "");
	taskbarRButtonUp = GetxRCLine(name, "OnRightClickUp", "");
	taskbarRButtonDown = GetxRCLine(name, "OnRightClickDown", "");
	taskbarMWheelDown = GetxRCLine(name, "OnWheelDown", "");
	taskbarMWheelUp = GetxRCLine(name, "OnWheelUp", "");
	taskbarEnter = GetxRCLine(name, "OnEnter", "");
	taskbarLeave = GetxRCLine(name, "OnLeave", "");
	taskbarMove = GetxRCLine(name, "OnMove", "");
	taskbarResize = GetxRCLine(name, "OnResize", "");
	taskbarAdd = GetxRCLine(name, "OnAdd", "");
	taskbarRemove = GetxRCLine(name, "OnRemove", "");
	taskbarSelect = GetxRCLine(name, "OnSelect", "");

	taskbarLineAdd = GetxRCLine(name, "OnLineAdd", "");
	taskbarLineRemove = GetxRCLine(name, "OnLineRemove", "");

	showIcon = GetxRCBool(name, "ShowIcon", true);
	showText = GetxRCBool(name, "ShowText", true);
	showToolTip = GetxRCBool(name, "ShowTooltip", true);

	useHoverState = GetxRCBool(name, "ShowHoverState");

	loadInBox = GetxRCBool(name, "LoadInBox");
	startHidden = GetxRCBool(name, "StartHidden");
	
	if (GetxRCBool(name, "FocusOnEnter"))
		focusOnEnter = 1;
	else
		focusOnEnter = 0;

	leftBorder = GetxRCInt(name, "LeftBorder", 0, 0);
	topBorder = GetxRCInt(name, "TopBorder", 0, 0);
	rightBorder = GetxRCInt(name, "RightBorder", 0, 0);
	bottomBorder = GetxRCInt(name, "BottomBorder", 0, 0);

	if (!refresh)
	{
		actualwidth = GetxRCDimen(name, "Width", screenX, screenX);
		actualheight = GetxRCDimen(name, "Height", 32, screenY);
	}

	int old_lines_rows = lines_rows;

	lines_rows = GetxRCInt(name, "Lines", 1, 1);
	maxlines_rows = GetxRCInt(name, "MaxLines", 100, 1);
	if (lines_rows > maxlines_rows)
		lines_rows = maxlines_rows;

	if ( refresh )
	{
		if ( old_lines_rows-lines_rows != 0 )
		{
			lines_rowsAdded = lines_rowsAdded+(old_lines_rows-lines_rows);
			if ( lines_rowsAdded < 0 )
				lines_rowsAdded = 0;
		}
	}

	maxTaskWidth = GetxRCInt(name, "ButtonMaxWidth", screenX, 0);
	maxTaskHeight = GetxRCInt(name, "ButtonMaxHeight", screenY, 0);

	wrapCount = GetxRCInt(name, "WrapCount", -1);
	if ( wrapCount < -1 || wrapCount == 0 )
		wrapCount = -1;
	if ( wrapCount == -1 && lines_rows > 1 )
		wrapCount = 5;

	if ( refresh )
	{
		if ( wrapCount == -1 )
		{
			lines_rowsAdded = 0;
		}
		else
		{
			int lines = backupExportedButtonCount/wrapCount;
			if ( backupExportedButtonCount%wrapCount != 0 )
				lines++;

			if ( lines > lines_rows+lines_rowsAdded )
				lines_rowsAdded = lines-lines_rows;
			else if ( lines <= lines_rows )
				lines_rowsAdded = 0; 
		}
	}

	layout = GetxRCNamedValue(name, "Layout", layoutValues, 0);

	direction = GetxRCNamedValue(name, "ButtonDirection", directionValues, 1);
	wrapDirection = GetxRCNamedValue(name, "ButtonWrapDirection", wrapdirectionValues, 1);

	if ( layout == 0 )
	{
		if ( direction == 3 )
			direction = 1;
		else if ( direction == 4 )
			direction = 2;
		if ( wrapDirection == 3 )
			wrapDirection = 1;
		else if ( wrapDirection == 4 )
			wrapDirection = 2;
	}
	else if ( layout == 1 )
	{
		if ( direction == 1 )
			direction = 3;
		else if ( direction == 2 )
			direction = 4;
		if ( wrapDirection == 1 )
			wrapDirection = 3;
		else if ( wrapDirection == 2 )
			wrapDirection = 4;
	}

	xSpacing = GetxRCInt(name, "ButtonxSpacing", 0);
	ySpacing = GetxRCInt(name, "ButtonySpacing", 0);

	if ( box )
		transparencyMode = true;
	else
		transparencyMode = GetxRCBool(name, "TransparencyMode");

	if (refresh)
	{
		if (buttontexture)
			delete buttontexture;
		if (buttonicon)
			delete buttonicon;
		if (buttonfont)
			delete buttonfont;
		if (taskbartexture)
			delete taskbartexture;

		taskbartexture = new TaskbarTexture();
		buttontexture = new ButtonTexture(name);
		buttonicon = new ButtonIcon(name);
		buttonfont = new ButtonFont(name);
	}
	taskbartexture->configure(name, transparencyMode);
	buttontexture->configure(name, transparencyMode);
	buttonicon->configure(name);
	buttonfont->configure(name);

	if ( transparencyMode && !taskbartexture->isTransparencyNeeded() && !buttontexture->isTransparencyNeeded() )
		transparencyMode = false;

	dropTimeOut = GetxRCInt(name, "DropDelay", 0, 0, 5000);
	timerRefresh = GetxRCInt(name, "UpdateInterval", 1000, 100);
	flashRefresh = GetxRCInt(name, "FlashInterval", timerRefresh, 100);

	if (!refresh)
	{
		if (box == NULL)
		{
			actualx = GetxRCCoord(name, "X", 0, screenX);
			actualy = GetxRCCoord(name, "Y", 0, screenY);
		}
		else
		{
			RECT rc;
			GetClientRect(box, &rc);

			actualx = GetxRCCoord(name, "X", 0, rc.right-rc.left);
			actualy = GetxRCCoord(name, "Y", 0, rc.bottom-rc.top);
		}
		
		reposition(3, actualx, actualy, actualwidth, actualheight);
	}

	if (!refresh)
		EnumWindows(Taskbar::EnumWindowsProc, (LPARAM) this);

	if (!refresh)
	{
		ToolTips = CreateWindow(TOOLTIPS_CLASS, NULL, TTS_ALWAYSTIP,
								0, 0, 0, 0,
								NULL, NULL, hInstance, NULL);

		SetWindowPos( ToolTips, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);

		SendMessage(ToolTips, TTM_SETMAXTIPWIDTH, 0, max(screenX/3, 500));
		
		if (ToolTips != 0)
		{
			TOOLINFO ti;
			RECT clientRect;

			GetClientRect(hWnd, &clientRect);
			ti.cbSize = sizeof(TOOLINFO);
			ti.uFlags = TTF_SUBCLASS;
			ti.hwnd = hWnd;
			ti.uId = 0;
			ti.rect = clientRect;
			ti.hinst = hInstance;
			ti.lpszText = 0;
			ti.lParam = 0;

			SendMessage(ToolTips, TTM_ADDTOOL, 0, LPARAM(&ti));
		}

		updateHint(hWnd, "", true);
	}

	alphaTransparency = GetxRCInt(name, "AlphaTransparency", 255, 0, 255);

	if ( !GetxRCBool(name, "UseSolidColors") && !GetxRCString(name, "Image", "").empty() )
		alphaMap = GetxRCBool(name, "AlphaMap");
	else
		alphaMap = false;

	char step[32], time[32];
	char *alphabuffers[] = {step, time};

	string fade = GetxRCLine(name, "CustomAlphaFade", "");
	
	int numTokens = 0;
	numTokens = LCTokenize( fade.c_str(), alphabuffers, 2, NULL);

	if (fade.empty())
		alphaFade = false;
	else if (numTokens == 1)
	{
		alphaFadeStep = atoi( step );
		if (alphaFadeStep < 1 || alphaFadeStep > 255)
			alphaFadeStep = 2;
		alphaFade = true;
	}
	else if (numTokens == 2)
	{
		alphaFadeStep = atoi( step );
		if (alphaFadeStep < 1 || alphaFadeStep > 255)
			alphaFadeStep = 2;
		alphaFadeDelay = atoi( time );
		if (alphaFadeDelay < 1 || alphaFadeDelay > 500)
			alphaFadeDelay = 10;
		alphaFade = true;
	}

	if (!alphaFade)
		alphaFade = GetxRCBool(name, "AlphaFade");

	if ( IsOS(OS_2KXP) && !refresh && (alphaTransparency != 255 || alphaFade || alphaMap))
	{
		ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

		if (alphaFade)
			UpdateWindowTransparent(hWnd, 0);
		else
			UpdateWindowTransparent(hWnd, alphaTransparency);
	}
	else if ( !IsOS(OS_2KXP) && !refresh)
	{
		alphaTransparency = 255;
		alphaFade = false;
		alphaMap = false;
	}

	if (!refresh)
	{
		setAlwaysOnTop( GetxRCBool(name, "AlwaysOnTop") );
	}
	else
	{
		CreateLayout();
	}
}

void Taskbar::SwitchToButton( LPCSTR direction )
{
	HWND current = NULL;
	HWND prev = NULL;
	HWND next = NULL;

	if (lastActive != NULL)
		current = lastActive;

	if (current == NULL || taskbuttons.size() == 0)
		return;

	for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
	{
		if((*tbit)->hButtonTask == current)
		{
			if ( stricmp(direction, "next") == NULL )
			{
				++tbit;
				if(tbit != taskbuttons.end())
					next = (*tbit)->hButtonTask;
				else
					next = (*taskbuttons.begin())->hButtonTask;	
				
				if( next != NULL)
					SelectWindow(next);
			}
			else if ( stricmp(direction, "prev") == NULL )
			{
				if (tbit == taskbuttons.begin())
					prev = (*(taskbuttons.end()-1))->hButtonTask;
				else
				{
					--tbit;
					prev = (*tbit)->hButtonTask;
				}
				
				if ( prev != NULL)
					SelectWindow(prev);
			}
			break;
		}
	}
}
	
//------------------------------------------------------------
//Set Parent Box (nearly never needed/called!)
//------------------------------------------------------------

void Taskbar::setBox(HWND newparent)
{
	if (box == newparent)
		return;

	if (alwaysOnTop)
		setAlwaysOnTop(false);
	
	box = newparent;

	ModifyStyleEx(hWnd, WS_EX_LAYERED, 0);
	
	ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
	SetParent(hWnd, newparent);

	transparencyMode = true;
}

void Taskbar::setAlwaysOnTop(bool setAlwaysOnTop)
{
	if (box)
		return;
	
	alwaysOnTop = setAlwaysOnTop;
	
	if(hWnd)
	{
		ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
		SetParent(hWnd, setAlwaysOnTop ? 0 : GetLitestepDesktop());
		ModifyStyle(hWnd, WS_CHILD, WS_POPUP);
		
		SetWindowPos(hWnd, setAlwaysOnTop ? HWND_TOPMOST : HWND_NOTOPMOST,
			0, 0, 0, 0,
			SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOSIZE );
	}
}

void Taskbar::setAlpha(int newAlphaTransparency)
{
	if (box)
		return;

	if ( newAlphaTransparency >= 0 && newAlphaTransparency <= 255 && IsOS(OS_2KXP) )
	{
		if (!layeredFlag)
		{
			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, 0);
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

			ModifyStyleEx(hWnd, 0, WS_EX_LAYERED);

			ModifyStyle(hWnd, WS_POPUP, WS_CHILD);
			SetParent(hWnd, alwaysOnTop ? 0 : GetLitestepDesktop());
			ModifyStyle(hWnd, WS_CHILD, WS_POPUP);

			layeredFlag = true;
		}

		alphaTransparency = newAlphaTransparency;

		UpdateWindowTransparent(hWnd, newAlphaTransparency);
	}
}

void Taskbar::UpdateWindowTransparent(HWND hWnd, int factor)
{
	/* First, see if we can get the API call we need. If we've tried
    * once, we don't need to try again. */
	if (!alphaInitialized)
	{
		HMODULE hDLL = LoadLibrary ("user32");

		pSetLayeredWindowAttributes = 
					(PSLWA) GetProcAddress(hDLL, "SetLayeredWindowAttributes");

		pUpdateLayeredWindow = 
					(PULW) GetProcAddress(hDLL, "UpdateLayeredWindow");

		alphaInitialized = TRUE;
   
		if (pSetLayeredWindowAttributes == NULL || pUpdateLayeredWindow == NULL)
		{
		    ModifyStyleEx(hWnd, WS_EX_LAYERED, 0 );

			alphaTransparency = 255;
			alphaMap = false;
			alphaFade = false;

			return;
		}
	}

	/* Now, we need to set the 'layered window attributes'. This
	 * is where the alpha values get set. */
  	if (alphaMap)
	{
		//------------------------------------------------------------------
		RECT rectPosition;
		GetWindowRect(hWnd, &rectPosition);
		int x = rectPosition.left;
		int y = rectPosition.top;
		int width = (rectPosition.right-rectPosition.left);
		int height = (rectPosition.bottom-rectPosition.top);

		// keep a cached rendition of the background
		if(!backgroundBitmap)
		{
			if(!backgroundDC)
				backgroundDC = CreateCompatibleDC(NULL);
			
			BITMAPINFO alpha;
			ZeroMemory( &alpha.bmiHeader, sizeof(BITMAPINFOHEADER) );
			alpha.bmiHeader.biWidth=width;      // Set size you need
			alpha.bmiHeader.biHeight=height;    // Set size you need
			alpha.bmiHeader.biPlanes=1;
			alpha.bmiHeader.biBitCount=32;            // Can be 8, 16, 32 bpp or other number
			alpha.bmiHeader.biSizeImage=0;
			alpha.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
			alpha.bmiHeader.biClrUsed= 0;
			alpha.bmiHeader.biClrImportant= 0;
			VOID *pvBits;

			backgroundBitmap = CreateDIBSection( backgroundDC,
										&alpha,
										DIB_RGB_COLORS,
										&pvBits,
										NULL,
										0 );

			SelectObject(backgroundDC, backgroundBitmap);
			
			//Normal Background Painting
			taskbartexture->apply(backgroundDC, 0, 0, width, height, true);
	
			if (displayButtonType >= 0)
			{
				for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
					(*tbit)->onPaint( backgroundDC );
			}
			else
			{
				for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
					(*tbit)->onPaint( backgroundDC );
			}

			//Prepare Combined Image
			BITMAP bm;
			GetObject(backgroundBitmap, sizeof(BITMAP), &bm);

			BYTE *pPixel = (LPBYTE)bm.bmBits;

			for (int j = 0; j < bm.bmHeight * bm.bmWidth; ++j)
			{
				//Convert Text to Full Opacity
				if (pPixel[3] == 0)
					pPixel[3] = 255;

				//Remove Prepared Alpha (1 back to 0)
				if (pPixel[3] == 1)
					pPixel[3] = 0;

				pPixel+= 4;
			}
		}

		POINT ptWindowScreenPosition = {x, y};
		SIZE szWindow = {width, height};

		POINT ptSrc = {0,0};

		BLENDFUNCTION blendPixelFunction = { AC_SRC_OVER, 0, factor, AC_SRC_ALPHA };

		pUpdateLayeredWindow(hWnd, NULL, &ptWindowScreenPosition, &szWindow, backgroundDC, &ptSrc, 0, &blendPixelFunction, ULW_ALPHA );
		//------------------------------------------------------------------*/
	}
	else
		pSetLayeredWindowAttributes(hWnd, 0, factor, LWA_ALPHA);
}

//------------------------------------------------------------
//Show/Hide
//------------------------------------------------------------
void Taskbar::show()
{
	if(hWnd)
	{
		if (alphaFade && !visible)
		{
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);

			KillTimer(hWnd, TIMER_ALPHAFADEOUT);
			SetTimer(hWnd, TIMER_ALPHAFADEIN, alphaFadeDelay, NULL);
		}
		else
			ShowWindow(hWnd, SW_SHOWNOACTIVATE);

		visible = true;

		repaint();
	}
}

void Taskbar::hide()
{
	if(hWnd)
	{
		if (alphaFade && visible)
		{
			UpdateAlpha = alphaTransparency;

			KillTimer(hWnd, TIMER_ALPHAFADEIN);
			SetTimer(hWnd, TIMER_ALPHAFADEOUT, alphaFadeDelay, NULL);
		}
		else
			ShowWindow(hWnd, SW_HIDE);

		visible = false;

		if ( box != NULL )
			SendMessage(box, LM_EXTERNALDRAWNOTIFY, 0, 0 );
	}
}

//------------------------------------------------------------
//Move/Resize/Reposition
//------------------------------------------------------------
//Type == 1 -> Move
//Type == 2 -> Resize
//Type == 3 -> Reposition

void Taskbar::reposition(int type, int newx, int newy, int newwidth, int newheight, int steps, int time, bool repositionBy)
{
	RECT rectPosition;
	GetWindowRect(hWnd, &rectPosition);

	if (steps > 0 && time > 0)
	{
		if (time > 50)
			time = 50;

		if (repositionBy)
		{
			if (type != 2)
			{
				newx = rectPosition.left + newx;
				newy = rectPosition.top + newy;
			}
			if (type != 1)
			{
				newwidth = (rectPosition.right-rectPosition.left) + newwidth;
				newheight = (rectPosition.bottom-rectPosition.top) + newheight;
			}
		}

		int counter = 0;
		int deltax = newx - rectPosition.left;
		int deltay = newy - rectPosition.top;
		int deltasizex = newwidth - (rectPosition.right-rectPosition.left);
		int deltasizey = newheight - (rectPosition.bottom-rectPosition.top);

		bool growx;
		bool growy;
		bool growsizex;
		bool growsizey;
		
		if (deltax < 0)
			growx = false;
		else
			growx = true;
		if (deltay < 0)
			growy = false;
		else
			growy = true;
		
		if (deltasizex < 0)
			growsizex = false;
		else
			growsizex = true;
		if (deltasizey < 0)
			growsizey = false;
		else
			growsizey = true;

		div_t div_result;
		div_result = div( deltax, steps );
		deltax = div_result.quot;
		div_result = div( deltay, steps );
		deltay = div_result.quot;

		div_result = div( deltasizex, steps );
		deltasizex = div_result.quot;
		div_result = div( deltasizey, steps );
		deltasizey = div_result.quot;

		if (deltax == 0 && growx)
			deltax = 1;
		else if (deltax == 0 && !growx)
			deltax = -1;
		if (deltay == 0 && growy)
			deltay = 1;
		else if (deltay == 0 && !growy)
			deltay = -1;

		if (deltasizex == 0 && growsizex)
			deltasizex = 1;
		else if (deltasizex == 0 && !growsizex)
			deltasizex = -1;
		if (deltasizey == 0 && growsizey)
			deltasizey = 1;
		else if (deltasizey == 0 && !growsizey)
			deltasizey = -1;

		int dynx = rectPosition.left;
		int dyny = rectPosition.top;
		int dynwidth = (rectPosition.right-rectPosition.left);
		int dynheight = (rectPosition.bottom-rectPosition.top);

		int stopper = 0;
		while (counter <= steps)
		{
			stopper = 0; // Stop if x, y, width and height are finished, stopper = 4
			dynx = dynx+deltax;
			if (growx && dynx > newx)
			{
				dynx = newx;
				stopper++;
			}
			else if (!growx && dynx < newx)
			{
				dynx = newx;
				stopper++;
			}

			dyny = dyny+deltay;
			if (growy && dyny > newy)
			{
				dyny = newy;
				stopper++;
			}
			else if (!growy && dyny < newy)
			{
				dyny = newy;
				stopper++;
			}

			dynwidth = dynwidth+deltasizex;
			if (growsizex && dynwidth > newwidth)
			{
				dynwidth = newwidth;
				stopper++;
			}
			else if (!growsizex && dynwidth < newwidth)
			{
				dynwidth = newwidth;
				stopper++;
			}

			dynheight = dynheight+deltasizey;
			if (growsizey && dynheight > newheight)
			{
				dynheight = newheight;
				stopper++;
			}
			else if (!growsizey && dynheight < newheight)
			{
				dynheight = newheight;
				stopper++;
			}
				
			SetWindowPos(hWnd, 0, dynx, dyny, dynwidth, dynheight,
					SWP_NOACTIVATE | (type == 2 ? SWP_NOMOVE : 0) | (type == 1 ? SWP_NOSIZE : 0 ) | SWP_NOZORDER);
				
			if (box != NULL)
				SendMessage(box, WM_PAINT, 0, 0);

			SendMessage(hWnd, WM_PAINT, 0, 0);

			if (stopper == 4)
				counter = steps;

			counter++;

			Sleep(time);
		}

		if ( type == 1 )
		{
			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
					SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
		}
		else if ( type == 2 )
		{
			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
					SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
		}
		else if ( type == 3 )
		{
			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
					SWP_NOACTIVATE | SWP_NOZORDER);
		}
	}
	else
	{
		if (repositionBy)
		{
			if (type != 2)
			{
				newx = rectPosition.left + newx;
				newy = rectPosition.top + newy;
			}
			if (type != 1)
			{
				newwidth = (rectPosition.right-rectPosition.left) + newwidth;
				newheight = (rectPosition.bottom-rectPosition.top) + newheight;
			}
		}

		if ( type == 1 )
		{
			SetWindowPos(hWnd, 0, newx, newy, 0, 0,
					SWP_NOACTIVATE | SWP_NOSIZE | SWP_NOZORDER);
		}
		else if ( type == 2 )
		{
			SetWindowPos(hWnd, 0, 0, 0, newwidth, newheight,
					SWP_NOACTIVATE | SWP_NOMOVE | SWP_NOZORDER);
		}
		else if ( type == 3 )
		{
			SetWindowPos(hWnd, 0, newx, newy, newwidth, newheight,
					SWP_NOACTIVATE | SWP_NOZORDER);
		}
	}
}

void Taskbar::updateRegion()
{
	if ( !alphaMap )
	{
		RECT rectPosition;
		GetWindowRect(hWnd, &rectPosition);

		HDC tmpDC = CreateCompatibleDC(NULL);
		HBITMAP tmpBmp = CreateBitmap((rectPosition.right-rectPosition.left), (rectPosition.bottom-rectPosition.top), 1, bitsperpixel, NULL);
		HGDIOBJ tmpObj = SelectObject(tmpDC, tmpBmp);

		if ( backgroundDC && backgroundBitmap )
			BitBlt(tmpDC, 0, 0, (rectPosition.right-rectPosition.left), (rectPosition.bottom-rectPosition.top), backgroundDC, 0, 0, SRCCOPY);
		else
			onPaint(tmpDC);
					
		SelectObject(tmpDC, tmpObj);

		HRGN region = BitmapToRegion(tmpBmp, RGB(255, 0, 255), 0, 0, 0);
		HRGN windowRgn = CreateRectRgn(0, 0, 0, 0);
		CombineRgn(windowRgn, region, NULL, RGN_COPY);

		SetWindowRgn(hWnd, windowRgn, true);	

		DeleteObject(windowRgn);
		DeleteObject(region);
		DeleteObject(tmpBmp);
		DeleteObject(tmpObj);
		DeleteDC(tmpDC);
	}
	else
	{
		HRGN windowRgn = CreateRectRgn(0, 0, actualwidth, actualheight);
		SetWindowRgn(hWnd, windowRgn, true);
		DeleteObject(windowRgn);
	}
}

//------------------------------------------------------------
//Mouse Related Actions (Clicks, Enter, Leave, Move, ...)
//------------------------------------------------------------

void Taskbar::onLButtonDown(int x, int y)
{
	TaskbarButton *button = ButtonAtPoint( x, y );

	if ( button != NULL )
	{
		if ( actionsOnDownButton )
			button->onButtonClick( x, y, lButtonAction );
		else
		{
			if ( button->currentState == STATE_SELECTED )
				minimizeAtAllCosts = true;
		}

		movecapture = button;
	}
	else if ( button == NULL && !taskbarLButtonDown.empty() )
	{
		LSExecute(hWnd, taskbarLButtonDown.c_str(), SW_SHOWNORMAL);
	}
}

void Taskbar::onLButtonUp(int x, int y)
{	
	TaskbarButton *button = ButtonAtPoint( x, y );

	if ( button != NULL && movecapture != NULL && button == movecapture )
	{
		if ( !actionsOnDownButton )
			button->onButtonClick( x, y, lButtonAction, minimizeAtAllCosts );
	}
	//Drag 'n Drop Positioning
	else if ( autoSort == 0 && button != NULL && movecapture != NULL && movecapture != button && displayButtonType < 0)
	{	
		vector<TaskbarButton *> tmptaskbuttons;
		
		for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
		{
			if ( *tbit == button )
			{
				if ( button->position < movecapture->position )
				{
					tmptaskbuttons.insert( tmptaskbuttons.end(), (movecapture) );
					tmptaskbuttons.insert( tmptaskbuttons.end(), (button) );
				}
				else if ( button->position > movecapture->position )
				{
					tmptaskbuttons.insert( tmptaskbuttons.end(), (button) );
					tmptaskbuttons.insert( tmptaskbuttons.end(), (movecapture) );
				}
			}
			else if ( *tbit != movecapture )
				tmptaskbuttons.insert( tmptaskbuttons.end(), (*tbit) );
		}
	
		taskbuttons = tmptaskbuttons;
		tmptaskbuttons.clear();
		movecapture = NULL;
		CreateLayout();
	}
	else if ( button == NULL && !taskbarLButtonUp.empty() )
	{
		LSExecute(hWnd, taskbarLButtonUp.c_str(), SW_SHOWNORMAL);
	}

	movecapture = NULL;

	minimizeAtAllCosts = false;
}

void Taskbar::onMButtonDown(int x, int y)
{
	TaskbarButton *button = ButtonAtPoint( x, y );

	if ( button != NULL )
	{
		if ( actionsOnDownButton )
			button->onButtonClick( x, y, mButtonAction );
		else
		{
			if ( button->currentState == STATE_SELECTED )
				minimizeAtAllCosts = true;
		}

		clickcapture = button;
	}
	else if ( button == NULL && !taskbarMButtonDown.empty() )
	{
		LSExecute(hWnd, taskbarMButtonDown.c_str(), SW_SHOWNORMAL);
	}
}

void Taskbar::onMButtonUp(int x, int y)
{
	TaskbarButton *button = ButtonAtPoint( x, y );

	if ( button != NULL && clickcapture != NULL && button == clickcapture )
	{
		if ( !actionsOnDownButton )
			button->onButtonClick( x, y, mButtonAction, minimizeAtAllCosts );
	}
	else if ( button == NULL && !taskbarMButtonUp.empty() )
	{
		LSExecute(hWnd, taskbarMButtonUp.c_str(), SW_SHOWNORMAL);
	}

	minimizeAtAllCosts = false;
	clickcapture = NULL;
}

void Taskbar::onRButtonDown(int x, int y)
{
	TaskbarButton *button = ButtonAtPoint( x, y );

	if ( button != NULL )
	{
		if ( actionsOnDownButton )
			button->onButtonClick( x, y, rButtonAction );
		else
		{
			if ( button->currentState == STATE_SELECTED )
				minimizeAtAllCosts = true;
		}

		clickcapture = button;
	}
	else if ( button == NULL && !taskbarRButtonDown.empty() )
	{
		LSExecute(hWnd, taskbarRButtonDown.c_str(), SW_SHOWNORMAL);
	}
}

void Taskbar::onRButtonUp(int x, int y)
{
	TaskbarButton *button = ButtonAtPoint( x, y );

	if ( button != NULL && clickcapture != NULL && button == clickcapture )
	{
		if ( !actionsOnDownButton )
			button->onButtonClick( x, y, rButtonAction, minimizeAtAllCosts );
	}
	else if ( button == NULL && !taskbarRButtonUp.empty() )
	{
		LSExecute(hWnd, taskbarRButtonUp.c_str(), SW_SHOWNORMAL);
	}
	
	minimizeAtAllCosts = false;
	clickcapture = NULL;
}

void Taskbar::onWheelDown(int x, int y)
{
	if(!taskbarMWheelDown.empty())
		LSExecute(hWnd, taskbarMWheelDown.c_str(), SW_SHOWNORMAL);

	if (focusOnEnter == 1)
	{
		SetForegroundWindow(hWnd);
	}
}

void Taskbar::onWheelUp(int x, int y)
{
	if(!taskbarMWheelUp.empty())
		LSExecute(hWnd, taskbarMWheelUp.c_str(), SW_SHOWNORMAL);

	if (focusOnEnter == 1)
	{
		SetForegroundWindow(hWnd);
	}
}

void Taskbar::onMouseEnter()
{
	if(!taskbarEnter.empty())
		LSExecute(hWnd, taskbarEnter.c_str(), SW_SHOWNORMAL);

	if (focusOnEnter == 1)
	{
		SetForegroundWindow(hWnd);
	}
}

void Taskbar::onMouseLeave()
{
	if(!taskbarLeave.empty())
		LSExecute(hWnd, taskbarLeave.c_str(), SW_SHOWNORMAL);
}

void Taskbar::onMouseMove(int x, int y)
{
	if(!mouseInside)
	{
		mouseInside = true;
		onMouseEnter();
		SetTimer(hWnd, TIMER_MOUSETRACK, 250, 0);
	}
}

void Taskbar::repaint()
{
	if(hWnd)
	{
		if(backgroundDC && backgroundBitmap)
		{
			backgroundBitmap = (HBITMAP) SelectObject(backgroundDC, backgroundBitmap);
			DeleteObject(backgroundBitmap);
			backgroundBitmap = 0;
		}

		if ( !alphaMap )
			InvalidateRect(hWnd, NULL, false);
	}

	if ( transparencyMode )
		updateRegion();

	//Attention!! Background MUST exist
	if ( alphaMap && IsOS(OS_2KXP) && taskbartexture && buttontexture && buttonicon && buttonfont )
		UpdateWindowTransparent(hWnd, alphaTransparency);

	if ( box != NULL )
		SendMessage(box, LM_EXTERNALDRAWNOTIFY, 0, 0 );
}

void Taskbar::onPaint(HDC hDC)
{	
	int width = actualwidth;
	int height = actualheight;
	// keep a cached rendition of the background
	if(!backgroundBitmap)
	{
		RECT rectPosition;
		GetWindowRect(hWnd, &rectPosition);

		int x = rectPosition.left;
		int y = rectPosition.top;
		width = (rectPosition.right-rectPosition.left);
		height = (rectPosition.bottom-rectPosition.top);

		rectPosition.left = 0;
		rectPosition.top = 0;
		rectPosition.right = width;
		rectPosition.bottom = height;

		if(!backgroundDC)
			backgroundDC = CreateCompatibleDC(hDC);

		backgroundBitmap = CreateCompatibleBitmap(hDC, width, height);
		SelectObject(backgroundDC, backgroundBitmap);

		//True Transparency
		if ( transparencyMode )
		{
			::SetBkColor(backgroundDC, RGB(255, 0, 255));
			::ExtTextOut(backgroundDC, 0, 0, ETO_OPAQUE, &rectPosition, NULL, 0, NULL);
		}
		//PrePaint Background (Fake Transparency)
		else if ( !transparencyMode )
		{
			//  =======================================================================
			//  Paint desktop on display DC and then into background buffer
			//  =======================================================================
			PaintDesktopEx(backgroundDC, 0, 0, width, height, x, y, FALSE);
		}
		
		//Normal Background Painting
		taskbartexture->apply(backgroundDC, 0, 0, width, height, transparencyMode);

		if (displayButtonType >= 0)
		{
			for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
				(*tbit)->onPaint( backgroundDC );
		}
		else
		{
			for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
				(*tbit)->onPaint( backgroundDC );
		}
	}

	// blt the double buffer to the display
	BitBlt(hDC, 0, 0, width, height, backgroundDC, 0, 0, SRCCOPY);
}

void Taskbar::onTimer(int timerID)
{
	switch(timerID)
	{	
		case TIMER_UPDATE:
		{
			if(!inCriticalProcess && visible)
			{
				bool needrepaint = false;
				for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
				{
					if ( (*tbit)->GetFlashing() || (*tbit)->onUpdate(false) || (*tbit)->getCurrentState() != (*tbit)->currentState )
						needrepaint = true;
				}

				if ( (needrepaint && (displayButtonType >= 0 || autoSort != 0)) || VWMCurrentDesktopOnly )
					CreateLayout();
				else if (needrepaint)
					repaint();
			}
		}
		break;

		case TIMER_MOUSETRACK:
		{
			POINT pt;
			GetCursorPos(&pt);
			
			RECT rc;
			GetWindowRect(hWnd, &rc);

			if(!PtInRect(&rc, pt))
			{
				KillTimer(hWnd, TIMER_MOUSETRACK);

				mouseInside = false;
				onMouseLeave();	
				updateHint(hWnd, "", false);
				PrefixedLSSetVariable(name, "HoverTask", "");
				if (useHoverState)
				{
					for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
						(*tbit)->useHover = false;

					repaint();
				}
				capture = NULL;
				break;
			}

			ScreenToClient(hWnd, &pt);
			TaskbarButton *button = ButtonAtPoint( pt.x, pt.y );

			if ( button && button != capture )
			{
				if (showToolTip)
					updateHint(hWnd, (char *) button->tooltip.c_str(), false);
				PrefixedLSSetVariable(name, "HoverTask", button->GetWindowCaption());
				if (useHoverState)
				{
					for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
						(*tbit)->useHover = false;

					button->useHover = true;
					repaint();	
				}
			}
			else if (button == NULL && capture != NULL)
			{
				if (showToolTip)
					updateHint(hWnd, "", false);
				PrefixedLSSetVariable(name, "HoverTask", "");
				if (useHoverState)
				{
					for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
						(*tbit)->useHover = false;

					repaint();
				}
			}
			capture = button;
		}
		break;

		case TIMER_ALPHAFADEIN:
		{
			if (UpdateAlpha <= alphaTransparency-alphaFadeStep)
			{
				UpdateWindowTransparent(hWnd, UpdateAlpha);
				UpdateAlpha += alphaFadeStep;
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEIN);
				if (UpdateAlpha != alphaTransparency)
					UpdateWindowTransparent(hWnd, alphaTransparency);
			}
		}
		break;

		case TIMER_ALPHAFADEOUT:
		{
			if (UpdateAlpha >= alphaFadeStep)
			{
				UpdateWindowTransparent(hWnd, UpdateAlpha);
				UpdateAlpha -= alphaFadeStep;
			}
			else
			{
				KillTimer(hWnd, TIMER_ALPHAFADEOUT);
				UpdateAlpha = 0;
				ShowWindow(hWnd, SW_HIDE);
			}
		}
		break;

		case TIMER_DROPTIMEOUT:
		{
			KillTimer(hWnd, TIMER_DROPTIMEOUT);

			POINT pt;
			GetCursorPos( &pt );
			ScreenToClient( hWnd, &pt );

			TaskbarButton* button = ButtonAtPoint( pt.x, pt.y );
			if( button && button == dropcapture )
			{
				if (button->currentState != STATE_SELECTED)
					SwitchToThisWindow( button->hButtonTask, SW_RESTORE );
			}

			dropcapture = NULL;
		}
		break;
	}
}

void Taskbar::onWindowPosChanging(WINDOWPOS *windowPos)
{
	if ( (windowPos->flags & SWP_NOZORDER) == 0 )
	{
		if ( !afterClass.empty() )
		{
			HWND before = FindWindowEx2(afterClass, afterTitle);
			if ( before != NULL )
			{
				if ( windowPos->hwndInsertAfter != before )
					windowPos->hwndInsertAfter = before;
			}
			else if ( !alwaysOnTop )
				windowPos->flags = windowPos->flags | SWP_NOZORDER;
		}
		else if ( !alwaysOnTop )
			windowPos->flags = windowPos->flags | SWP_NOZORDER;
	}
}

void Taskbar::onMove(int x, int y)
{
	PrefixedLSSetVariable(name, "CurrentX", IntToString( x ));
	PrefixedLSSetVariable(name, "CurrentY", IntToString( y ));
	
	if ( createLayoutLocked )
		return;

	actualx = x;
	actualy = y;

	if ( !transparencyMode && (taskbartexture->isTransparencyNeeded() || !buttontexture->isTransparencyNeeded()) )
		repaint();

	if( !taskbarMove.empty() )
		LSExecute(hWnd, taskbarMove.c_str(), SW_SHOWNORMAL);
}

void Taskbar::onSize(int width, int height)
{
	PrefixedLSSetVariable(name, "CurrentWidth", IntToString( width ));
	PrefixedLSSetVariable(name, "CurrentHeight", IntToString( height ));
	
	if ( createLayoutLocked )
		return;

	actualwidth = width;
	actualheight = height;
	
	CreateLayout();

	//Set new Rect for ToolTip Tracker
	if ( showToolTip && ToolTips )
		updateHint(hWnd, "", true);
	
	if( !taskbarResize.empty() )
		LSExecute(hWnd, taskbarResize.c_str(), SW_SHOWNORMAL);
}

BOOL Taskbar::onWindowMessage( UINT nMessage, WPARAM wParam, LPARAM lParam, LRESULT &lResult )
{
	lResult = 0;

	switch ( nMessage )
	{
		case LM_EXTERNALDRAW:
		{
			if ( visible )
			{
				HDC desthDC = (HDC) wParam;

				if ( alphaMap || alphaTransparency != 255 )
				{
					BLENDFUNCTION blendFunction = { AC_SRC_OVER, 0, alphaTransparency, (alphaMap ? AC_SRC_ALPHA : 0) };

					AlphaBlend(desthDC, actualx, actualy, actualwidth, actualheight, backgroundDC, 0, 0, actualwidth, actualheight, blendFunction);
				}
				else
				{
					TransparentBltLS(desthDC, actualx, actualy, actualwidth, actualheight, backgroundDC, 0, 0, RGB(255, 0, 255));
				}
			}
			return true;
		}

		case WM_CLOSE:
		{
			lResult = 0;
			return true;
		}
	
		case WM_DESTROY:
		{
			if(box)		// FIXME: this is the old workaround
			{
				if (hWnd && ToolTips)
				{
					updateHint(hWnd,"", false, true);
					DestroyWindow(ToolTips);
					ToolTips = 0;
				}

				if (hWnd)
				{
					KillTimer(hWnd, TIMER_MOUSETRACK);
					KillTimer(hWnd, TIMER_UPDATE);
					KillTimer(hWnd, TIMER_ALPHAFADEIN);
					KillTimer(hWnd, TIMER_ALPHAFADEOUT);
					KillTimer(hWnd, TIMER_DROPTIMEOUT);
				}

				box = NULL;
				hide();
			}
				
			hWnd = NULL;	// FIXME: dirty hack
			
			return false;
		}

		case WM_LBUTTONDOWN:
		{
			onLButtonDown( LOWORD(lParam), HIWORD(lParam));
			return true;
		}

		case WM_LBUTTONUP:
		{
			onLButtonUp( LOWORD(lParam), HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONDOWN:
		{
			onMButtonDown( LOWORD(lParam), HIWORD(lParam));
			return true;
		}

		case WM_MBUTTONUP:
		{
			onMButtonUp( LOWORD(lParam), HIWORD(lParam));
			return true;
		}

		case WM_RBUTTONDOWN:
		{
			onRButtonDown( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return true;
		}

		case WM_RBUTTONUP:
		{
			onRButtonUp( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return true;
		}

		case WM_MOUSEWHEEL:
		{
			if ((short)(HIWORD(wParam)) < 0)
			{
				onWheelDown((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
				return true;
			}
			else
			{
				onWheelUp((int) (short) LOWORD(lParam), (int) (short) HIWORD(lParam));
				return true;
			}
		}

		case WM_MOUSEMOVE:
		{
			onMouseMove( LOWORD( lParam ), HIWORD( lParam ) );
			return true;
		}

		case LM_LSSELECT:
		{
			onWindowActivated( (HWND) wParam, FALSE );
			return true;
		}

		case LM_REDRAW:
		{
			onRedraw( (HWND) wParam, (BOOL) lParam ); 
			return true;
		}

		case LM_WINDOWACTIVATED:
		{
			onWindowActivated( (HWND) wParam, (BOOL) lParam );
			return true;
		}

		case LM_WINDOWCREATED:
		{
			if ( IsAppWindow( (HWND) wParam ) )
			{
				onWindowCreated( (HWND) wParam );
				CreateLayout();
			}
			return true;
		}

		case LM_WINDOWDESTROYED:
		{
			onWindowDestroyed( (HWND) wParam );
			CreateLayout();
			return true;
		}

		if ( !alphaMap || !IsOS(OS_2KXP) )
		{
			case WM_PAINT:
			{
				if (!inCriticalProcess && visible)
				{
					PAINTSTRUCT ps;

					HDC hDC;

					hDC = BeginPaint( hWnd, &ps );

					onPaint( hDC );

					EndPaint( hWnd, &ps );
				}

				return true;
			}
		}

		case WM_WINDOWPOSCHANGING:
		{
			onWindowPosChanging( (WINDOWPOS *) lParam );
			return false;
		}

		case WM_MOVE:
		{
			onMove( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return true;
		}

		case WM_SIZE:
		{
			onSize( (int) (short) LOWORD( lParam ), (int) (short) HIWORD( lParam ) );
			return true;
		}

		case WM_TIMER:
		{
			onTimer((int) wParam);
			return true;
		}

		case LM_UPDATEBG:
		{
			PaintDesktopEx(0, 0, 0, 0, 0, 0, 0, TRUE);

			repaint();

			return false;
		}

		case WM_DISPLAYCHANGE:
		{
			return true;
		}

		case WM_SETTINGCHANGE:
		{
			PostMessage(hWnd, LM_UPDATEBG, 0, 0);
			return false;
		}

		case MOUSEISDRAGGING:
		{
			if ( dropcapture == NULL )
			{
				POINT pt;
				pt.x = GET_X_LPARAM( lParam );
				pt.y = GET_Y_LPARAM( lParam );

				ScreenToClient( hWnd, &pt );

				TaskbarButton* button = ButtonAtPoint( pt.x, pt.y );
				if( button )
				{
					if ( dropTimeOut != 0 )
					{
						dropcapture = button;
						SetTimer(hWnd, TIMER_DROPTIMEOUT, dropTimeOut, 0);
					}
					else
					{
						if (button->currentState != STATE_SELECTED)
							SwitchToThisWindow( button->hButtonTask, SW_RESTORE );
					}
				}
			}
		}
	}

	return false;
}

TaskbarButton *Taskbar::ButtonAtPoint( int x, int y )
{
	for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
	{
		if ( (*tbit)->HasPoint(x, y) )
			return (*tbit);
	}

	return NULL;
}

TaskbarButton *Taskbar::SearchForButtonByHandle( HWND search_hWnd )
{
	for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
	{
		if ( search_hWnd == (*tbit)->hButtonTask  )
			return (*tbit);
	}
	
	return NULL;
}

//Actually first tries to get Exe-Filename before using Class
TaskbarButton *Taskbar::SearchForButtonByClass( HWND search_hWnd )
{
	for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
	{
		if ( stricmp( (GetTaskFileName( search_hWnd )).c_str(), (*tbit)->classname.c_str()) == NULL )
			return (*tbit);
	}

	return NULL;
}

BOOL WINAPI Taskbar::EnumWindowsProc(HWND hWnd, LPARAM lParam)
{
	Taskbar *taskbar = (Taskbar *) lParam;
	if (taskbar && IsAppWindow(hWnd))
	{
		//Grouping
		TaskbarButton *searchbutton = taskbar->SearchForButtonByClass( hWnd );
		if (searchbutton != NULL && taskbar->useGrouping)
		{
			searchbutton->AddToGroup( hWnd );
		}
		else
		{
			// Not on general RemoveList, not on custom RemoveList and on MatchingList (default is true)
			if ( !IsOnTaskbarList(hWnd, taskbar->RemoveList, false) && !IsOnTaskbarList(hWnd, taskbar->RemoveMatchingList, false) && IsOnTaskbarList(hWnd, taskbar->AddMatchingList, true) )
			{
				//DoubleCheck, if not already there
				if ( taskbar->SearchForButtonByHandle(hWnd) == NULL )
				{
					TaskbarButton *button = new TaskbarButton( taskbar, hWnd );
					taskbar->taskbuttons.insert(taskbar->taskbuttons.end(), button);

					++nButtonsAddedAtStartup;
				}
			}
		}
	}
	return TRUE;
}

LRESULT WINAPI Taskbar::windowProcedure(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	Taskbar *taskbar = (Taskbar *) GetWindowLong(hWnd, 0);

	if (message == WM_NCCREATE)
	{
		LPCREATESTRUCT pcs = (LPCREATESTRUCT) lParam;
		taskbar = (Taskbar *) pcs->lpCreateParams;

		taskbar->hWnd = hWnd;
		SetWindowLong(hWnd, 0, (LONG) taskbar);
	}

	if (taskbar)
	{
		LRESULT lResult;

		if (taskbar->onWindowMessage(message, wParam, lParam, lResult))
			return lResult;
	}

	return DefWindowProc(hWnd, message, wParam, lParam);
}

void Taskbar::CreateLayout()
{
	if ( createLayoutLocked )
		return;

	inCriticalProcess = true;

	RECT rectPosition;
	GetWindowRect(hWnd, &rectPosition);

	int x = rectPosition.left;
	int y = rectPosition.top;
	int width = (rectPosition.right-rectPosition.left);
	int height = (rectPosition.bottom-rectPosition.top);

	int buttonCount = 0;

	displayedtaskbuttons.clear();

	for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
	{
		if ( displayButtonType == -2 )
		{
			if ( (*tbit)->GetFlashing() && IsValidWindow((*tbit)->hButtonTask, VWMCurrentDesktopOnly) )
			{
				buttonCount++;
				(*tbit)->position = buttonCount;
				displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
			}
			else
				(*tbit)->setButtonLayout( 2000, 2000, 0, 0 );
		}
		else if (GetState((*tbit)->hButtonTask) == STATE_NORMAL 
				&& (displayButtonType < 0 || displayButtonType == 0 || displayButtonType == 3 || displayButtonType == 4))
		{
			if ( IsValidWindow((*tbit)->hButtonTask, VWMCurrentDesktopOnly) )
			{
				buttonCount++;
				(*tbit)->position = buttonCount;
				displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
			}
			else
				(*tbit)->setButtonLayout( 2000, 2000, 0, 0 );
		}
		else if (GetState((*tbit)->hButtonTask) == STATE_SELECTED 
			&& (displayButtonType < 0 || displayButtonType == 1 || displayButtonType == 3 || displayButtonType == 5))
		{
			if ( IsValidWindow((*tbit)->hButtonTask, VWMCurrentDesktopOnly) )
			{
				buttonCount++;
				(*tbit)->position = buttonCount;
				displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
			}
			else
				(*tbit)->setButtonLayout( 2000, 2000, 0, 0 );
		}
		else if (GetState((*tbit)->hButtonTask) == STATE_MINIMIZED 
				&& (displayButtonType < 0 || displayButtonType == 2 || displayButtonType == 4 || displayButtonType == 5))
		{
			if ( IsValidWindow((*tbit)->hButtonTask, VWMCurrentDesktopOnly) )
			{
				buttonCount++;
				(*tbit)->position = buttonCount;
				displayedtaskbuttons.insert(displayedtaskbuttons.end(), (*tbit));
			}
			else
				(*tbit)->setButtonLayout( 2000, 2000, 0, 0 );
		}
		else
			(*tbit)->setButtonLayout( 2000, 2000, 0, 0 );
	}

	PrefixedLSSetVariable(name, "CurrentButtonCount", IntToString( buttonCount ));

	if (buttonCount <= 0)
	{
		PrefixedLSSetVariable(name, "CurrentLines", IntToString( lines_rows ));

		displayedtaskbuttons.clear();

		inCriticalProcess = false;

		repaint();

		backupExportedButtonCount = buttonCount;

		if ( hideIfEmpty && !autoHidden )
		{
			autoHidden = true;
			hide();
		}

		return;
	}
	else if ( autoHidden )
	{
		autoHidden = false;
		show();
	}

	if ( autoSort > 0 )
		SortTasks(displayedtaskbuttons);

	if (layout == 0)
	{
		//Start Horizontal Version
		//-----------------------------------------------------------------------
		int maxavailableWidth = max( width-leftBorder-rightBorder, 0 );
	
		int fireLineCount = 0;

		if ( wrapCount != -1 )
		{
			if ( addLock != buttonCount && buttonCount/wrapCount >= lines_rows+lines_rowsAdded && buttonCount%wrapCount == 1 && backupExportedButtonCount < buttonCount )
			{
				addLock = buttonCount;
				lines_rowsAdded++;
				if ( !taskbarLineAdd.empty() && lines_rows+lines_rowsAdded <= maxlines_rows )
				{	
					fireLineCount = 1;
				}
				else if ( lines_rows+lines_rowsAdded > maxlines_rows)
				{
					lines_rowsAdded--;
					wrapCount++;
				}
			}
			else if ( removeLock != buttonCount && lines_rowsAdded >= 1 && buttonCount/wrapCount >= lines_rows && buttonCount%wrapCount == 0 && backupExportedButtonCount > buttonCount)
			{
				removeLock = buttonCount;
				lines_rowsAdded--;
				if ( !taskbarLineRemove.empty() )
					fireLineCount = -1;
			}
			else
			{
				addLock = -1;
				removeLock = -1;
			}
		}

		int realLines = min(lines_rows+lines_rowsAdded, maxlines_rows);

		PrefixedLSSetVariable(name, "CurrentLines", IntToString( realLines ));

		if ( fireLineCount == -1 )
		{
			LSExecute(hWnd, taskbarLineRemove.c_str(), SW_SHOWNORMAL);

			inCriticalProcess = false;

			backupExportedButtonCount = buttonCount;

			return;
		}
		else if ( fireLineCount == 1 )
		{
			LSExecute(hWnd, taskbarLineAdd.c_str(), SW_SHOWNORMAL);

			inCriticalProcess = false;

			backupExportedButtonCount = buttonCount;

			return;
		}

		if ( buttonCount > 1 )
		{
			if ( wrapCount != -1 )
				maxavailableWidth = maxavailableWidth - (wrapCount-1) * xSpacing;
			else
				maxavailableWidth = maxavailableWidth - (buttonCount-1) * xSpacing;
		}

		int buttonWidth = 0;
		if ( wrapCount != -1 )
		{
			if ( maxavailableWidth%wrapCount == 0 )
				buttonWidth = maxavailableWidth/wrapCount;
			else
				buttonWidth = (maxavailableWidth/wrapCount)+1;
		}
		else
		{
			if ( maxavailableWidth/buttonCount > maxTaskWidth )
			{
				buttonWidth = maxTaskWidth;
			}
			else
			{
				if ( maxavailableWidth%buttonCount == 0 )
					buttonWidth = maxavailableWidth/buttonCount;
				else
					buttonWidth = (maxavailableWidth/buttonCount)+1;
			}
		}
		int buttonHeight = min( (height - topBorder - bottomBorder - (realLines - 1) * ySpacing)/realLines, maxTaskHeight);
	
		int x = 0;
		int y = 0;

		int buttoncounter = 1;
		int hcounter = 1;
		int vcounter = 1;
		// set placement for each button
		for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
		{
			if (direction == 1) //Right
				x = leftBorder + (hcounter-1) * (buttonWidth + xSpacing);
			else if (direction == 2) // Left
				x = (width - rightBorder) - hcounter * buttonWidth - (hcounter-1) * xSpacing;
			
			if ( wrapDirection == 1) //Bottom
				y = topBorder + (vcounter-1) * (buttonHeight + ySpacing);
			else if ( wrapDirection == 2) //Top
				y = (height - bottomBorder) - vcounter * buttonHeight - (vcounter-1) * ySpacing;

			//Resize Button, if required
			if (x+buttonWidth > width-rightBorder)
				(*tbit)->setButtonLayout( x, y, width-rightBorder-x, buttonHeight );
			else if (x < leftBorder)
				(*tbit)->setButtonLayout( leftBorder, y, buttonWidth-(leftBorder-x), buttonHeight );
			else
				(*tbit)->setButtonLayout( x, y, buttonWidth, buttonHeight );

			if ( wrapCount != -1 && buttoncounter%wrapCount == 0 )
			{
				hcounter = 0;
				vcounter++;
			}
			hcounter++;
			buttoncounter++;
		}
		//-----------------------------------------------------------------------
		//End Horizontal Version
	}
	else if (layout == 1)
	{
		//Start Vertical Version
		//-----------------------------------------------------------------------
		int maxavailableHeight = max( height-topBorder-bottomBorder, 0 );
	
		int fireLineCount = 0;

		if ( wrapCount != -1 )
		{
			if ( addLock != buttonCount && buttonCount/wrapCount >= lines_rows+lines_rowsAdded && buttonCount%wrapCount == 1 && backupExportedButtonCount < buttonCount )
			{
				addLock = buttonCount;
				lines_rowsAdded++;
				if ( !taskbarLineAdd.empty() && lines_rows+lines_rowsAdded <= maxlines_rows )
				{	
					fireLineCount = 1;
				}
				else if ( lines_rows+lines_rowsAdded > maxlines_rows)
				{
					lines_rowsAdded--;
					wrapCount++;
				}
			}
			else if ( removeLock != buttonCount && lines_rowsAdded >= 1 && buttonCount/wrapCount >= lines_rows && buttonCount%wrapCount == 0 && backupExportedButtonCount > buttonCount)
			{
				removeLock = buttonCount;
				lines_rowsAdded--;
				if ( !taskbarLineRemove.empty() )
					fireLineCount = -1;
			}
			else
			{
				addLock = -1;
				removeLock = -1;
			}
		}

		int realLines = min(lines_rows+lines_rowsAdded, maxlines_rows);

		PrefixedLSSetVariable(name, "CurrentLines", IntToString( realLines ));

		if ( fireLineCount == -1 )
		{
			LSExecute(hWnd, taskbarLineRemove.c_str(), SW_SHOWNORMAL);
			
			inCriticalProcess = false;

			backupExportedButtonCount = buttonCount;

			return;
		}
		else if ( fireLineCount == 1 )
		{
			LSExecute(hWnd, taskbarLineAdd.c_str(), SW_SHOWNORMAL);
			
			inCriticalProcess = false;

			backupExportedButtonCount = buttonCount;

			return;
		}

		if ( buttonCount > 1 )
		{
			if ( wrapCount != -1 )
				maxavailableHeight = maxavailableHeight - (wrapCount-1) * ySpacing;
			else
				maxavailableHeight = maxavailableHeight - (buttonCount-1) * ySpacing;
		}

		int buttonHeight = 0;
		if ( wrapCount != -1 )
		{
			if ( maxavailableHeight%wrapCount == 0 )
				buttonHeight = maxavailableHeight/wrapCount;
			else
				buttonHeight = (maxavailableHeight/wrapCount)+1;
		}
		else
		{
			if ( maxavailableHeight/buttonCount > maxTaskHeight )
			{
				buttonHeight = maxTaskHeight;
			}
			else
			{
				if ( maxavailableHeight%buttonCount == 0 )
					buttonHeight = maxavailableHeight/buttonCount;
				else
					buttonHeight = (maxavailableHeight/buttonCount)+1;
			}
		}
		int buttonWidth = min( (width - leftBorder - rightBorder - (realLines - 1) * xSpacing)/realLines, maxTaskWidth);

		int x = 0;
		int y = 0;

		int buttoncounter = 1;
		int hcounter = 1;
		int vcounter = 1;
		// set placement for each button
		for (TaskbarButtonIterator tbit = displayedtaskbuttons.begin(); tbit != displayedtaskbuttons.end(); ++tbit)
		{	
			if ( direction == 3 ) //bottom
				y = topBorder + (vcounter-1) * (buttonHeight + ySpacing);
			else if ( direction == 4)  //top
				y = (height - bottomBorder) - vcounter * buttonHeight - (vcounter-1) * ySpacing;
	
			if ( wrapDirection == 3 ) //right
				x = leftBorder + (hcounter-1) * (buttonWidth + xSpacing);
			else if ( wrapDirection == 4 ) //left
				x = (width - rightBorder) - hcounter * buttonWidth - (hcounter-1) * xSpacing;

			//Make Smaller if required
			if (y+buttonHeight > height-bottomBorder)
				(*tbit)->setButtonLayout( x, y, buttonWidth, height-bottomBorder-y );
			else if (y < topBorder)
				(*tbit)->setButtonLayout( x, topBorder, buttonWidth, buttonHeight-(topBorder-y) );
			else
				(*tbit)->setButtonLayout( x, y, buttonWidth, buttonHeight );

			if ( wrapCount != -1 && buttoncounter%wrapCount == 0 )
			{
				hcounter++;
				vcounter = 0;
			}
			vcounter++;
			buttoncounter++;
		}
		//-----------------------------------------------------------------------
		//End Vertical Version
	}
	inCriticalProcess = false;
	
	repaint();

	backupExportedButtonCount = buttonCount;
}

void Taskbar::onRedraw( HWND hApplication, bool fFlash )
{
	if ( useGrouping )
	{
		TaskbarButton *searchbutton = SearchForButtonByClass( hApplication );
		if (searchbutton)
		{
			if(GetActiveWindow() != searchbutton->hButtonTask)
			{
				searchbutton->SetFlashing(fFlash);
				if (fFlash)
				{
					KillTimer(hWnd, TIMER_UPDATE);
					SetTimer(hWnd, TIMER_UPDATE, flashRefresh, NULL);
				}
			}
		}
	}
	else
	{
		TaskbarButton *searchbutton = SearchForButtonByHandle( hApplication );
		if (searchbutton)
		{
			if(GetActiveWindow() != searchbutton->hButtonTask)
			{
				searchbutton->SetFlashing(fFlash);
				if (fFlash)
				{
					KillTimer(hWnd, TIMER_UPDATE);
					SetTimer(hWnd, TIMER_UPDATE, flashRefresh, NULL);
				}
			}
		}
	}
}

void Taskbar::onWindowActivated( HWND hApplication, BOOL fFullScreen )
{
	if ( useGrouping )
	{
		TaskbarButton *searchbutton = SearchForButtonByClass( hApplication );
		if (searchbutton)
		{
			searchbutton->hButtonTask = hApplication;
			if (searchbutton->GetFlashing())
			{
				KillTimer(hWnd, TIMER_UPDATE);
				SetTimer(hWnd, TIMER_UPDATE, timerRefresh, NULL);
			}
			searchbutton->SetFlashing(FALSE);
			searchbutton->currentState = GetState(searchbutton->hButtonTask);
		}
	}
	else
	{
		TaskbarButton * searchbutton = SearchForButtonByHandle( hApplication );
		if (searchbutton)
		{
			if (searchbutton->GetFlashing())
			{
				KillTimer(hWnd, TIMER_UPDATE);
				SetTimer(hWnd, TIMER_UPDATE, timerRefresh, NULL);
			}
			searchbutton->SetFlashing(FALSE);
			searchbutton->currentState = GetState(searchbutton->hButtonTask);	
		}
	}
	
	repaint();
}

void Taskbar::onWindowCreated( HWND hApplication )
{
	bool reallyAdd = false;
	//Grouping
	TaskbarButton *searchbutton = SearchForButtonByClass( hApplication );
	if (searchbutton != NULL && useGrouping)
	{
		searchbutton->AddToGroup( hApplication );
	}
	else
	{
		// Not on RemoveList and on MatchingList (default is true, if empty)
		if ( !IsOnTaskbarList(hApplication, RemoveList, false) && !IsOnTaskbarList(hApplication, RemoveMatchingList, false) && IsOnTaskbarList(hApplication, AddMatchingList, true) )
		{
			//DoubleCheck, if not already there
			if ( SearchForButtonByHandle(hApplication) == NULL )
			{
				TaskbarButton *button = new TaskbarButton( this, hApplication );
				taskbuttons.insert(taskbuttons.end(), button);

				if(!taskbarAdd.empty() && displayButtonType < 0)
				{
					PrefixedLSSetVariable(name, "CurrentButtonCount", IntToString( backupExportedButtonCount+1 ));
							
					LSExecute(hWnd, taskbarAdd.c_str(), SW_SHOWNORMAL);
				}
			}
		}
	}
}

void Taskbar::onWindowDestroyed( HWND hApplication )
{
	//Grouping
	if (useGrouping)
	{
		for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit)
		{
			if ( (*tbit)->hwndgrouplist.size() > 1 && (*tbit)->RemoveFromGroup( hApplication ) )
				return;
		}
	}

	for (TaskbarButtonIterator tbit = taskbuttons.begin(); tbit != taskbuttons.end(); ++tbit )
	{	
		if ( (*tbit)->hButtonTask == hApplication )
		{
			delete (*tbit);
			taskbuttons.erase( tbit );

			if(!taskbarRemove.empty() && displayButtonType < 0)
			{
				PrefixedLSSetVariable(name, "CurrentButtonCount", IntToString( backupExportedButtonCount-1 ));
				LSExecute(hWnd, taskbarRemove.c_str(), SW_SHOWNORMAL);
			}
			break;
		}
	}
}

void Taskbar::updateHint(HWND hWnd, LPSTR caption, bool updatesize, bool remove)
{
	if (ToolTips == 0)
		return;	
	
	if (!remove)
	{
		TOOLINFO ti;
		RECT clientRect;

		GetClientRect(hWnd, &clientRect);
		ti.cbSize = sizeof(TOOLINFO);
		ti.uFlags = TTF_SUBCLASS;
		ti.hwnd = hWnd;
		ti.uId = 0;
		ti.rect = clientRect;
		ti.hinst = hInstance;
		ti.lpszText = caption;
		ti.lParam = 0;

		SendMessage(ToolTips, TTM_UPDATETIPTEXT, 0, LPARAM(&ti));

		if (updatesize)
			SendMessage(ToolTips, TTM_NEWTOOLRECT, 0, LPARAM(&ti));
	}
	else
	{
		TOOLINFO ti;
		RECT emptyRect = {0, 0, 0, 0};

		ti.cbSize = sizeof(TOOLINFO);
		ti.uFlags = 0;
		ti.hwnd = hWnd;
		ti.uId = 0;
		ti.rect = emptyRect;
		ti.hinst = hInstance;
		ti.lpszText = NULL;
		ti.lParam = 0;

		SendMessage(ToolTips, TTM_DELTOOL, 0, LPARAM(&ti));
	}
}

void Taskbar::SortTasks(vector<TaskbarButton *> &taskbuttons)
{
	std::sort( taskbuttons.begin(), taskbuttons.end(), CompareTasks );
}